// Class CBreakThrough64, Version 1.0, Feb-2013  

#include "CEngineDefinitions.h"
#include <intrin.h>	// POPCNT

#include "CBreakThrough64.h"


CBreakThrough64::CBreakThrough64()
{

	BreakThrough64_Init() ;

}


CBreakThrough64::~CBreakThrough64()
{
}


BITBOARD CBreakThrough64::OPPathWhiteNext(int iPos)
{
	BITBOARD bbOPPath, bbPosition ;

	if ( iPos < 0 || iPos >= 64 )
		return ( 0 ) ;

	bbOPPath	= 0 ;
	bbPosition	= (BITBOARD)1<<iPos ;

	if ( bbPosition & DFLD_ALL ) {

		bbOPPath  = bbPosition ;
		bbOPPath |= OPPathWhiteNext(iPos + DIRECTION_FL ) ;
		bbOPPath |= OPPathWhiteNext(iPos + DIRECTION_FR ) ;
	}

	return ( bbOPPath ) ;
}


BITBOARD CBreakThrough64::OPPathBlackNext(int iPos)
{
	BITBOARD bbOPPath, bbPosition ;

	if ( iPos < 0 || iPos >= 64 )
		return ( 0 ) ;

	bbOPPath	= 0 ;
	bbPosition	= (BITBOARD)1<<iPos ;

	if ( bbPosition & DFLD_ALL ) {

		bbOPPath  = bbPosition ;
		bbOPPath |= OPPathBlackNext(iPos + DIRECTION_BR ) ;
		bbOPPath |= OPPathBlackNext(iPos + DIRECTION_BL ) ;
	}

	return ( bbOPPath ) ;
}


void CBreakThrough64::BreakThrough64_Init(void)
{
	int i ;
	BITBOARD bbBitBoard ;


	for ( i=0; i<64; i++ ) {

		if ( DFLD_ALL & (BITBOARD)1<<i ) {

			m_bbOPPathWhite[i]  = OPPathWhiteNext(i + DIRECTION_FL) ;
			m_bbOPPathWhite[i] |= OPPathWhiteNext(i + DIRECTION_FR) ;

			m_bbOPPathBlack[i]  = OPPathBlackNext(i + DIRECTION_BR) ;
			m_bbOPPathBlack[i] |= OPPathBlackNext(i + DIRECTION_BL) ;

		} else {

			m_bbOPPathWhite[i]  = 0 ;
			m_bbOPPathBlack[i]  = 0 ;
		}
	}

// Calculate Row
	for ( i=0; i<64; i++ ) {

		bbBitBoard = (BITBOARD)1<<i ;

		if ( bbBitBoard & DFLD_ALL ) {

			if ( bbBitBoard & DFLD_ROW0 )		// Row 0
				m_iRow[i] = 0 ;
			else if ( bbBitBoard & DFLD_ROW1 )	// Row 1
				m_iRow[i] = 1 ;
			else if ( bbBitBoard & DFLD_ROW2 )	// Row 2
				m_iRow[i] = 2 ;
			else if ( bbBitBoard & DFLD_ROW3 )	// Row 3
				m_iRow[i] = 3 ;
			else if ( bbBitBoard & DFLD_ROW4 )	// Row 4
				m_iRow[i] = 4 ;
			else if ( bbBitBoard & DFLD_ROW5 )	// Row 5
				m_iRow[i] = 5 ;
			else if ( bbBitBoard & DFLD_ROW6 )	// Row 6
				m_iRow[i] = 6 ;
			else if ( bbBitBoard & DFLD_ROW7 )	// Row 7
				m_iRow[i] = 7 ;
			else if ( bbBitBoard & DFLD_ROW8 )	// Row 8
				m_iRow[i] = 8 ;
			else if ( bbBitBoard & DFLD_ROW9 )	// Row 9
				m_iRow[i] = 9 ;
		} else

			m_iRow[i] = -1 ;
	}
}


int CBreakThrough64::WhiteBreakThrough(PTSTACK pTDepth, bool* pbBreakThrough, bool bBreakThroughMode)
{
	bool bKing ;
	int iCountWhite, iCountBlack, iOffset, iPly, iRow, iScore ;
	DWORD iBitSet ;
	BITBOARD bbWhiteMan0, bbWhiteMan, bbBlackMan, bbBlockMan, bbLockedMan, bbOPPathWhite, bbPosition ;

	pTDepth->bOutputBoard = false ;

	iScore			= 0 ;
	*pbBreakThrough	= false ;

	// Test on Open PromotionPath
	bbWhiteMan0 = pTDepth->bbField[BB_WHITEMAN] & DFLD_BREAKTHROUGH_WHITE ;	// Position 6 - 30

	if ( bbWhiteMan0 == 0 )
		return ( 0 ) ;	// No BreakThrough, Return Score 0

	bbWhiteMan	= bbWhiteMan0 ;
	bbBlackMan	= pTDepth->bbField[BB_BLACKMAN] ;
	bbLockedMan	= 0 ;

	iOffset = ( pTDepth->bTurn == true ? 0 : 64 ) ;
	bKing	= ( pTDepth->bbField[BB_BLACKKING]  ? true : false ) ;

	while ( bbWhiteMan ) {

		_BitScanForward64(&iBitSet, bbWhiteMan) ;

		bbPosition = (BITBOARD)1<<iBitSet ;

		bbWhiteMan ^= bbPosition ;

		bbOPPathWhite	= m_bbOPPathWhite[iBitSet] ;
		bbBlockMan		= bbOPPathWhite & bbBlackMan ;

		iCountBlack = (int)POPCNT64(bbBlockMan) ;
		iCountWhite = (int)POPCNT64(bbOPPathWhite & bbWhiteMan0) ;

		*pbBreakThrough = false ;

		if ( iCountBlack == 0 && iCountWhite >= 0 ) {

			if ( iCountWhite == 0 && ( bbPosition & DFLD15 ) && ( bbBlackMan & DFLD14 ) )	// Special Case
				iScore += 40 ;
			else
				*pbBreakThrough = true ;
		
		} else if ( iCountBlack == 1 && iCountWhite == 1 ) {

			// Special cases no Breakthrough
			if ( ALLPIECE( bbWhiteMan0, ( DFLD15 | DFLD20 ) ) && ALLPIECE( bbBlackMan, ( DFLD4 | DFLD19 ) ) )
				iScore += 30 ;
			else
				*pbBreakThrough = true ;

		} else if ( iCountBlack == 1 && iCountWhite > 1 ) {
			*pbBreakThrough = true ;
		
		} else if ( iCountBlack == 1 && iCountWhite == 0 ) {

			if ( bbNBWhite[iBitSet + iOffset] & bbBlackMan )
				*pbBreakThrough = true ;
			else if ( bbLockedMan & bbBlockMan )	// the Blocking Man is already Locked through another Attacker
				*pbBreakThrough = true ;
			else
				bbLockedMan |= bbBlockMan ;

		} else if ( iCountBlack == 2 && iCountWhite == 0 ) {		// 2 Black Man defend 1 White Man against a breakthrough.

			if ( ( bbPosition & DFLD_ROW2 ) && ( POPCNT64( bbOPPathWhite & bbBlackMan & DFLD_ROW0 ) == 2 ) )
				iScore += 40 ;
		}

		if ( *pbBreakThrough == true ) {

			if ( bBreakThroughMode == true )	// Return only bBreakThrough ( no Score ) 
				return ( 0 ) ;

			iRow	= m_iRow[iBitSet] ;
			iPly	= 2* iRow - ( pTDepth->bTurn == true ? 2 : 1 ) ;

			if ( bKing )
				iScore += ( m_iBreakThrough64Score[iPly] / 2 ) ;
			else
				iScore += m_iBreakThrough64Score[iPly] ;
		}

	}

	return ( iScore ) ;
}


bool CBreakThrough64::bWhiteBreakThrough(PTSTACK pTDepth)
{
	bool bBreakThrough ;
	int iScore ;

	iScore = WhiteBreakThrough(pTDepth, &bBreakThrough, true) ;

	return ( bBreakThrough ) ;
}


int CBreakThrough64::BlackBreakThrough(PTSTACK pTDepth, bool* pbBreakThrough, bool bBreakThroughMode)
{
	bool bKing ;
	int iCountWhite, iCountBlack, iOffset, iPly, iRow, iScore ;
	DWORD iBitSet ;
	BITBOARD bbWhiteMan, bbBlackMan, bbBlackMan0, bbBlockMan, bbLockedMan, bbOPPathBlack, bbPosition ;

	pTDepth->bOutputBoard = false ;

	iScore			= 0 ;
	*pbBreakThrough = false ;

	// Test on Open PromotionPath
	bbBlackMan0 = pTDepth->bbField[BB_BLACKMAN] & DFLD_BREAKTHROUGH_BLACK ;	// Position 21 - 45

	if ( bbBlackMan0 == 0 )
		return ( false ) ;

	bbBlackMan	= bbBlackMan0 ;
	bbWhiteMan	= pTDepth->bbField[BB_WHITEMAN] ;
	bbLockedMan	= 0 ;

	iOffset = ( pTDepth->bTurn == true ? 0 : 64 ) ;
	bKing	= ( pTDepth->bbField[BB_WHITEKING] ? true : false ) ;

	while ( bbBlackMan ) {

		_BitScanForward64(&iBitSet, bbBlackMan) ;

		bbPosition = (BITBOARD)1<<iBitSet ;

		bbBlackMan ^= bbPosition ;

		bbOPPathBlack	= m_bbOPPathBlack[iBitSet] ;
		bbBlockMan		= bbOPPathBlack & bbWhiteMan ;

		iCountWhite = (int)POPCNT64(bbBlockMan) ;
		iCountBlack = (int)POPCNT64(bbOPPathBlack & bbBlackMan0) ;

		*pbBreakThrough = false ;

		if ( iCountWhite == 0 && iCountBlack >= 0 ) {

			if ( iCountBlack == 0 && ( (BITBOARD)1<<iBitSet & DFLD36 ) && ( pTDepth->bbField[BB_WHITEMAN] & DFLD37 ) )	// Special Case
				iScore += 40 ;
			else
				*pbBreakThrough = true ;

		} else if ( iCountWhite == 1 && iCountBlack == 1 ) {

			// Special cases no Breakthrough
			if ( ALLPIECE( bbBlackMan0, ( DFLD31 | DFLD36 ) ) && ALLPIECE( bbWhiteMan, ( DFLD32 | DFLD47 ) ) )
				iScore += 30 ;
			else
				*pbBreakThrough = true ;

		
		} else if ( iCountWhite == 1 && iCountBlack > 1 ) {
			*pbBreakThrough = true ;
		
		} else if ( iCountWhite == 1 && iCountBlack == 0 ) {

			if ( bbNBBlack[iBitSet + iOffset] & bbWhiteMan )
				*pbBreakThrough = true ;
			else if ( bbLockedMan & bbBlockMan )	// the Blocking Man is already Locked through another Attacker
				*pbBreakThrough = true ;
			else
				bbLockedMan |= bbBlockMan ;

		} else if ( iCountWhite == 2 && iCountBlack == 0 ) {		// 2 White Man defend 1 Black Man against a breakthrough.

			if ( ( bbPosition & DFLD_ROW7 ) && ( POPCNT64( bbOPPathBlack & bbWhiteMan & DFLD_ROW9 ) == 2 ) )
				iScore += 40 ;
		}

		if ( *pbBreakThrough == true ) {

			if ( bBreakThroughMode == true )	// Return only bBreakThrough ( no Score ) 
				return ( 0 ) ;

			iRow	= m_iRow[iBitSet] ;
			iPly	= ( pTDepth->bTurn == true ? 17 : 16 ) - 2* iRow ;

			if ( bKing )
				iScore += ( m_iBreakThrough64Score[iPly] / 2 ) ;
			else
				iScore += m_iBreakThrough64Score[iPly] ;			

		}	
	}

	return ( iScore ) ;
}


bool CBreakThrough64::bBlackBreakThrough(PTSTACK pTDepth)
{
	bool bBreakThrough ;
	int iScore ;

	iScore = BlackBreakThrough(pTDepth, &bBreakThrough, true) ;

	return ( bBreakThrough ) ;
}
